/*!
 * [NOTE] Cannot make the window a floating window.
 *
 * If we would make Selective's window a B_FLOATING_WINDOW, then it is 
 * always clear to the user in which Tracker window the selection is being 
 * made, because this window is still active (i.e. it has a yellow tab).
 *
 * Unfortunately, when Selective is invoked on the Desktop (instead of on
 * a Tracker window) it doesn't appear, because its window needs a Tracker
 * window as its parent. (Actually, as soon as the user selects a Tracker
 * window, Selective appears.) So, if we would make Selective a floating
 * window, the user would never be able to select stuff on the Desktop.
 */

// ***************************************************************************
// Implementation System Includes
// ***************************************************************************

#include <Box.h>
#include <Button.h>
#include <Directory.h>
#include <Messenger.h>
#include <Screen.h>
#include <Window.h>

#include <ctype.h>  // toupper()
#include <string.h>

// ***************************************************************************
// Implementation Project Includes
// ***************************************************************************

#include "Main.h"
#include "SelectiveWindow.h"

// ***************************************************************************
// Implementation Constant Definitions
// ***************************************************************************

/** The messages that the views send to the window. */
enum {
  MSG_SELECT = 1000,
  MSG_OK,
  MSG_CANCEL
};

// ***************************************************************************
// Implementation Type Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Variable Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Declarations
// ***************************************************************************

// ***************************************************************************
// Global Variable Definitions
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions: Lifecycle
// ***************************************************************************

// ===========================================================================
// Constructor
// ===========================================================================

tSelectiveWindow::tSelectiveWindow(entry_ref aDirRef, BMessage* apMessage) 
  : BWindow(BRect(0, 0, 1, 1),
            "Selective",
            B_TITLED_WINDOW,
            B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
{
  // Initialize attributes.
  mTrackerDirRef = aDirRef;
  mpTrackerMessage = apMessage;

  // Create the selection pattern entry field.
  mpPatternText = new BTextControl(BRect(0, 0, 0, 1),
          "Pattern",
          0,                           // No label.
          0,                           // No initial text.
          new BMessage(MSG_SELECT),
          B_FOLLOW_LEFT_RIGHT | B_FOLLOW_TOP,
          B_WILL_DRAW | B_FULL_UPDATE_ON_RESIZE | B_NAVIGABLE); 

  mpPatternText->SetModificationMessage(new BMessage(MSG_SELECT));
  mpPatternText->TextView()->SetMaxBytes(1000); 
  mpPatternText->ResizeToPreferred();

  // Create the "Ok" button.
  BButton* lpOkButton = new BButton(BRect(0, 1, 80, 1),
          "OkButton",
          "OK",
          new BMessage(MSG_OK),
          B_FOLLOW_RIGHT);
          
  lpOkButton->ResizeToPreferred();  

  // Create the "Cancel" button.
  BButton* lpCancelButton = new BButton(BRect(0, 1, 80, 1), 
          "CancelButton",
          "Cancel",
          new BMessage(MSG_CANCEL),
          B_FOLLOW_RIGHT);

  lpCancelButton->ResizeToPreferred();  

  // Now all views have been created, we can calculate how large the window
  // will be.
  tFloat lWidth  = 8
                 + lpCancelButton->Frame().Width()
                 + 8
                 + lpOkButton->Frame().Width()
                 + 8;
  
  tFloat lHeight = 8
                 + mpPatternText->Frame().Height()
                 + 8 
                 + lpCancelButton->Frame().Height()
                 + 8;

  // Resize the window to its final size.
  ResizeTo(lWidth, lHeight);  
  
  // Center the window on the screen.
  BRect lScreenRect = BScreen(this).Frame();
  MoveTo((lScreenRect.Width()-lWidth)/2, (lScreenRect.Height()-lHeight)/2);
  
  // Create the background box.
  BBox* lpBox  = new BBox(BRect(0, 0, lWidth, lHeight),
          B_EMPTY_STRING, 
          B_FOLLOW_ALL, 
          B_WILL_DRAW | B_FRAME_EVENTS | B_NAVIGABLE_JUMP, 
          B_PLAIN_BORDER); 

  AddChild(lpBox);

  // Add the selection pattern entry field to the background box.
  mpPatternText->MoveTo(8, 8);
  lpBox->AddChild(mpPatternText);
  mpPatternText->ResizeBy(lWidth - 16, 0);
  mpPatternText->MakeFocus(TRUE);

  // Add the "select" button to the background box.
  lpOkButton->MoveTo(
          lWidth  - lpOkButton->Frame().Width()  - 8, 
          lHeight - lpOkButton->Frame().Height() - 8);
  lpBox->AddChild(lpOkButton); 

  // Add the "cancel" button to the background box.
  lpCancelButton->MoveTo(
          8, 
          lHeight - lpCancelButton->Frame().Height() - 8);
  lpBox->AddChild(lpCancelButton); 
  
  // Finally show the window.
  Show(); 
}

// ***************************************************************************
// PUBLIC Member Function Definitions: Operators
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions: Operations
// ***************************************************************************

// ===========================================================================
// MessageReceived
// ===========================================================================

void
tSelectiveWindow::MessageReceived(BMessage* apMessage)
{
  // Figure out what kind of message was received.
  switch(apMessage->what) {
  
    // Make the selection. ---------------------------------------------------
    case MSG_SELECT: {
      
      // Do the funky stuff.
      Select();
      break;
    }

    // Accept the selection. -------------------------------------------------
    case MSG_OK: {
      
      // Commit suicide.
      PostMessage(B_QUIT_REQUESTED);
      break;
    }

    // Cancel the selection. -------------------------------------------------
    case MSG_CANCEL: {

      // Hide our window.
      Hide();
    
      // If the user already selected files, then we'll
      // simply select all of these files again.
      if (mpTrackerMessage->HasRef("refs") == TRUE) {
        mpPatternText->SetText("*");      
        Select();
        
      // Otherwise, we'll have to clear the selection.
      } else {
      
        // To do...
      }
      
      // Commit suicide.
      PostMessage(B_QUIT_REQUESTED);
      break;
    }

    // Let the original BWindow handle the message. --------------------------
    default: {
      BWindow::MessageReceived(apMessage);
      break;
    }
  }
}

// ***************************************************************************
// PUBLIC Member Function Definitions: Access
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions: Inquiry
// ***************************************************************************

// ***************************************************************************
// PROTECTED Member Function Definitions
// ***************************************************************************

// ***************************************************************************
// PRIVATE Member Function Definitions
// ***************************************************************************

// ===========================================================================
// Select
// ===========================================================================

void
tSelectiveWindow::Select()
{
  // Get a messenger to the Poses property of the
  // Tracker window that the user called us from.
  BMessenger lPoses;
  if (mpTrackerMessage->FindMessenger("TrackerViewToken", &lPoses) == B_OK) {
    if (lPoses.IsValid()) {
     
      /*!
       * [BUG] When no files match, the old selection isn't deleted.
       *
       * We need to tell the Poses property to clear its current
       * selection before making a new selection. This is necessary 
       * because the Tracker will not automatically clear the old 
       * selection if the new selection message doesn't contain any
       * entry_refs. Unfortunately, I don't seem to be able to delete 
       * the selection, no matter what I try.
       *
       * I cannot release version 1.2 before this bug is fixed!!!
       *
       * @date Saturday, 3 April 1999
       */

      // Clear the current selection.
      BMessage lGetReply;
      BMessage lGetMsg(B_GET_PROPERTY);
      lGetMsg.AddSpecifier("Selection");
      lPoses.SendMessage(&lGetMsg, &lGetReply);

      if (lGetReply.HasRef("result") == TRUE) {

        BBG_WRITE(("Found selection!"));

        BMessage lDeleteMsg(B_DELETE_PROPERTY);

        BMessage lDeleteSpecifier('sref');
        lDeleteSpecifier.AddString("property", "Selection");

        tInt32 lIndex = 0;
        entry_ref lCurrentRef;
        while (lGetReply.FindRef("result", lIndex, &lCurrentRef) == B_OK) {

          BBG_WRITE(("Adding entry_ref %s", lCurrentRef.name));

          lDeleteSpecifier.AddRef("refs", &lCurrentRef);
          ++lIndex;
        }

        lDeleteMsg.AddSpecifier(&lDeleteSpecifier);
        lDeleteMsg.PrintToStream();
        lPoses.SendMessage(&lDeleteMsg, &lGetReply);
        lGetReply.PrintToStream();
      }

      // Create a message that tells the Poses to make a new selection.
      BMessage lSelectionMsg(B_SET_PROPERTY);
      lSelectionMsg.AddSpecifier("Selection");

      // Did the user already select files?
      if (mpTrackerMessage->HasRef("refs") == TRUE) {

        tInt32 lIndex = 0;
        entry_ref lCurrentRef;
        while (mpTrackerMessage->FindRef("refs", 
                                         lIndex, 
                                         &lCurrentRef) == B_OK) {
          
          // If the current entry matches the wildcard pattern, then
          // we will add its entry_ref to the selection message.
          if (IsMatch(mpPatternText->Text(), lCurrentRef.name)) {
            lSelectionMsg.AddRef("data", &lCurrentRef);
          }

          // Get the next entry_ref from the message.          
          ++lIndex;
        }

      // No files selected yet.
      } else {
      
        // Loop through the entries inside the directory.
        BDirectory lDir(&mTrackerDirRef);
        entry_ref lCurrentRef;
        while (lDir.GetNextRef(&lCurrentRef) == B_OK) {

          // If the current entry matches the wildcard pattern, then
          // we will add its entry_ref to the selection message.
          if (IsMatch(mpPatternText->Text(), lCurrentRef.name)) {
            lSelectionMsg.AddRef("data", &lCurrentRef);
          }
        }
      }

      // Send the selection message to the Poses property.
      lPoses.SendMessage(&lSelectionMsg);
    }
  }
}

// ===========================================================================
// IsMatch
// ===========================================================================

bool
tSelectiveWindow::IsMatch(const tChar* apPattern, const tChar* apString)
{
  tInt32 lPatternIndex = 0;
  tInt32 lStringIndex = 0;
  size_t lPatternLength = strlen(apPattern);
  size_t lStringLength = strlen(apString);

  // Loop until we are finished with either the pattern or the string.
  while ((lPatternIndex < lPatternLength) 
  &&     (lStringIndex < lStringLength)) {
  
    // Figure out what character the string must match next.
    switch (apPattern[lPatternIndex]) {

      // Any character zero or more times.
      case '*': {
    
        // Add this to match any character once or more.
        // ++lStringIndex;

        // Skip the '*'.
        ++lPatternIndex;
        if (lPatternIndex < lPatternLength) {

          // Keep increasing the string index and calling ourselves
          // recursively until the rest of the string matches the pattern.
          while (lStringIndex < lStringLength) {
            if (IsMatch(apPattern+lPatternIndex, apString+lStringIndex)) {
              return true;
            }
            ++lStringIndex;          
          }
          return false;

        // The '*' was the last character in the pattern, 
        // so the rest of the string always matches.
        } else {
          return true;
        }
        break;
      }
      
      // The next character from the string may be any single character.
      case '?': {
        ++lPatternIndex;
        ++lStringIndex;
        break;
      }
      
      // The next character from the string must be exactly equal to the
      // next character from the pattern (not regarding case). 
      default: {
        if (toupper(apPattern[lPatternIndex]) 
        ==  toupper(apString[lStringIndex])) {
          ++lPatternIndex;
          ++lStringIndex;
        } else {
          return false;
        }
        break;
      }
    }
  }

  // Both indexes must have reached the end of their strings, 
  // otherwise there was no match.  
  if ((lPatternIndex < lPatternLength)
  ||  (lStringIndex < lStringLength)) {
    return false;
  }

  // If we made it this far, then the pattern matched.  
  return true;
}

// ***************************************************************************
// Helper Function Definitions
// ***************************************************************************

// ***************************************************************************
// Global Function Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Definitions
// ***************************************************************************
